.386p
locals

comment *

  This module contains the code for a "thunk" that will get called
  directly from the interrupt occuring.  The thunk prevents reentrancy
  by using a semaphore to direct subsequent interrupt calls to either
  the old vector (by chaining) or to an IRET.  The thunk also sets up a
  new stack (saving the old one), the size of which is controled by
  MAX_INTERRUPT_SIZE (default is 4k).  Interrupts are enabled so that
  additional ones can fire if need be, then the C language handler is
  called.  Upon return from the C handler, the thunk reverses the
  process and then either chains or IRET's.

  The thunk code is declared in the data segment so that we can
  dynamically allocate one for each interrupt that needs to be hooked.
  We also have to preform our own fixups to the code in the thunk, since
  we will be copying it to another chunk of memory.

*

MAX_INTERRUPT_SIZE EQU 1024d  ; 1024 dwords is the default for a 4k stack

_DATA segment dword public 'DATA' use32
	assume cs:_DATA

dummy	proc near

@@startThunk		label byte

	PUBLIC	__interrupt32Thunk

__interrupt32Thunk	label byte

  cli
  push ds
	mov    	ds, word ptr cs:[012345678h] ; this address is fixed up elseware
	PUBLIC	__I32TdsLoadLabel
__I32TdsLoadLabel	EQU $ - @@startThunk - 4
  cmp ds:[semaphore], 0
  jne dont_reenter

  inc ds:[semaphore]
  sti

	push	eax
	push	ebx
	push	ecx
	push	edx
	push	es
	push	esi
	push	edi
	push	ebp

  mov ds:[savess], ss
  mov ds:[savesp], esp
  push ds
  pop ss
  mov esp, offset newstkend

	db	0e8h      ; call the target C fuction
	dd	0         ; the address is fixed up in the helper code

	PUBLIC	__I32TcallLabel

__I32TcallLabel		EQU $ - @@startThunk - 4

  mov ss, ds:[savess]
  mov esp, ds:[savesp]

	pop	ebp
	pop	edi
	pop	esi
	pop	es
	pop	edx
	pop	ecx
	pop	ebx
	pop	eax

  cli
  dec ds:[semaphore]

dont_reenter:
 	cmp	dword ptr chainOffset, 0
	PUBLIC	__I32TchainCmp
__I32TchainCmp		EQU $ - @@startThunk - 5

	pop	ds
	jne	@@chain
@@iret:
	iret

@@chain:
	jmp	fword ptr [chainOffset]

	PUBLIC	__I32TchainLabel
__I32TchainLabel	EQU $ - @@startThunk - 4

	PUBLIC	__I32TchainOffset
__I32TchainOffset	EQU $ - @@startThunk
chainOffset	label	dword
	dd	0

	PUBLIC	__I32TchainSel
__I32TchainSel		EQU $ - @@startThunk
	dw	0

semaphore dd 0                       ; this prevents reentrancy
savess dw 0
savesp dd 0
newstk dd MAX_INTERRUPT_SIZE dup (0) ; default is a 4k stack
newstkend dd 0

	PUBLIC	__interrupt32ThunkSize
__interrupt32ThunkSize	EQU $ - @@startThunk
dummy	endp
_DATA ends
end

